        /*  DOS MCB compatible DOS memory manager for DOSRUN           *
         *  (C)1999 LGB, Gabor Lenart, lgb@vlug.vein.hu                *
         *  Can be distributed under the General Public License (GPL)  */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/mman.h>
#include "dosrun.h"
#include "memory.h"
#include "derrno.h"
#include "config_dosrun.h"


/* Well. After experiencing with varius DOS memory managment models, I discovered
   that the best way is to implement DOS compatible MCBs because its simplity and
   compatiblity with programs do direct MCB modification techniques.
   Most functions here requires `byte *' as input/output parameter or
   the linear address of memory data block (NOT THE MCB!). Size parameters
   are expected in 16 byte units.
   TODO: some (?) DOS versions stores program name in MCBs two. (8 bytes)
         Novell DOS uses only 7 bytes to be sure to put closer `\0' after string.
   BUGS: MCB corruption messages (if not casues by sily DOS application)
   indicates memory.c bug, so recompile with DEBUG_MEMORY_FUNCTION_PROBLEMS
   below, and send the output for us.
*/



byte *mcb_first;
static byte *memtop;             /* this is the only variable presents in BYTES, NOT 16 byte UNITS ! */

#define DEBUG_MEMORY_FUNCTION_PROBLEMS

#define START_ROM_AREA  0xC0000
#define END_ROM_AREA   0x100000


/* calling this function will protect 'rom' area : writing any data to this
   area will generate segmentation fault for dosrun. TODO: handle that
   sigsegv ... */
void dosmem_romprotect ( void )
{
  mprotect((void*)START_ROM_AREA,END_ROM_AREA-START_ROM_AREA+1,PROT_READ|PROT_EXEC);
}


/* Sets the owner of a memory block. */
void dosmem_setowner ( int addr , int owner )
{
  byte *p=(byte *)(addr-MCB_SIZE);
  MCB_X_OWNER(p)=owner;
}



/* Searchs for the largest free memory block */
int dosmem_largest ( void )
{
  byte *p=mcb_first;
  int msize=0;
#ifdef DEBUG_MEMORY_FUNCTION_PROBLEMS
  if (stddebug) fprintf(stddebug,"dosmem_largest enter\n");
#endif
  for (;;) {
    if (!MCB_X_OWNER(p)&&MCB_X_SIZE(p)>msize) msize=MCB_X_SIZE(p);
    if (MCB_X_TYPE(p)!='M') break; else p=MCB_NEXT(p);
  }
  if (MCB_X_TYPE(p)!='Z') {
#ifdef CONFIG_DEBUG_MCB_PROBLEMS
    if (stddebug) fprintf(stddebug,"MCB list corrupted (dosmem_largest) !!\n");
#endif
    return 0;
  }
#ifdef DEBUG_MEMORY_FUNCTION_PROBLEMS
  if (stddebug) fprintf(stddebug,"dosmem_largest leave : %04X\n",msize);
#endif
  return msize;
}



/* Searchs for free memory block based on required size and algorithm */
byte *dosmem_search ( int size , int mode )
{
  byte *p=mcb_first,*best=NULL;
  int bestremain=0;
#ifdef DEBUG_MEMORY_FUNCTION_PROBLEMS
  if (stddebug) fprintf(stddebug,"dosmem_search enter size=%04X mode=%d\n",size,mode);
#endif
  if (mode==SEARCH_BEST) bestremain=40960;   /* max 640K/16 */
  for (;;) {
    int remain=MCB_X_SIZE(p)-size;
    if (!MCB_X_OWNER(p)&&remain>=0)
      switch (mode) {
        case SEARCH_FIRST : return p;
        case SEARCH_LAST  : best=p;
	                    break;
        case SEARCH_BEST  : if (remain<bestremain) {
                              bestremain=remain;
       	                      best=p;
			    }
			    break;
	case SEARCH_WORST : if (remain>bestremain) {
	                      bestremain=remain;
			      best=p;
			    }
			    break;
      }
    if (MCB_X_TYPE(p)!='M') break; else p=MCB_NEXT(p);
  }
  if (MCB_X_TYPE(p)!='Z') {
#ifdef CONFIG_DEBUG_MCB_PROBLEMS
    if (stddebug) fprintf(stddebug,"MCB list corrupted (dosmem_search) !!\n");
    dosmem_dump();
#endif
    return NULL;
  }
#ifdef DEBUG_MEMORY_FUNCTION_PROBLEMS
  if (stddebug) fprintf(stddebug,"dosmem_search leave : %04X\n",((int)best+MCB_SIZE)>>4);
#endif
  return best;
}



/* Initializes vm86 memory and base MCB, the ONLY function requires size in bytes ! */
void dosmem_init ( int base , int ramsize )
{
  if (mmap(NULL,0x110000,PROT_READ|PROT_WRITE|PROT_EXEC,MAP_FIXED|MAP_PRIVATE|MAP_ANON,-1,0)) {
    perror("anonymous mmap()-ing 1M+64K failed");
    exit(1);
  }
  bzero(NULL,0x110000);
  memtop=(byte*)ramsize;
  mcb_first=(byte*)base;
  MCB_X_SIZE(mcb_first)=(ramsize-base-MCB_SIZE)>>4;
  MCB_X_OWNER(mcb_first)=0;
  MCB_X_TYPE(mcb_first)='Z';
}



/* Optimizes MCB chain after dosmem_free to avoid fragmented free blocks */
static int optimize_blocks ( void )
{
  byte *p=mcb_first,*q;
#ifdef DEBUG_MEMORY_FUNCTION_PROBLEMS
  if (stddebug) fprintf(stddebug,"optimize blocks enter\n");
  dosmem_dump();
#endif
  for (;;) {
    if (!MCB_X_OWNER(p)) {
      q=p;
      while (q&&!MCB_X_OWNER(q))
        if (MCB_X_TYPE(q)=='Z') q=NULL; else q=MCB_NEXT(q);
      if (!q) {
        MCB_X_TYPE(p)='Z';
	MCB_X_SIZE(p)=((int)memtop-(int)p-MCB_SIZE)>>4;
      } else if (p!=q)
        MCB_X_SIZE(p)=((int)q-(int)p-(MCB_SIZE<<1))>>4;
    }
    if (MCB_X_TYPE(p)!='M') break; else p=MCB_NEXT(p);
  }
  if (MCB_X_TYPE(p)!='Z') {
#ifdef CONFIG_DEBUG_MCB_PROBLEMS
    if (stddebug) fprintf(stddebug,"MCB list corrupted (optimize blocks) !!\n");
    dosmem_dump();
#endif
    return DBADMCB;
  }
#ifdef DEBUG_MEMORY_FUNCTION_PROBLEMS
  if (stddebug) fprintf(stddebug,"optimize blocks leave\n");
#endif
  return 0;
}



/* Splits a block into two blocks */
static byte *split_block ( byte *p , int size )
{
  byte *q=(byte*)((int)p+(size<<4)+MCB_SIZE);
#ifdef DEBUG_MEMORY_FUNCTION_PROBLEMS
  if (stddebug) fprintf(stddebug,"split blocks enter\n");
#endif
  MCB_X_SIZE(q) = MCB_X_SIZE(p)-size-MCB_PSIZE;
  MCB_X_TYPE(q) = MCB_X_TYPE(p);
  MCB_X_TYPE(p) = 'M';
  MCB_X_OWNER(q)= 0;
  MCB_X_SIZE(p) = size;
#ifdef DEBUG_MEMORY_FUNCTION_PROBLEMS
  if (stddebug) fprintf(stddebug,"split blocks leave\n");
#endif
  return q;
}



/* Joins to blocks */
static __inline__ void join_blocks ( byte *p )
{
#ifdef DEBUG_MEMORY_FUNCTION_PROBLEMS
  if (stddebug) fprintf(stddebug,"join blocks enter\n");
#endif
  MCB_X_TYPE(p)=MCB_X_TYPE(MCB_NEXT(p));
  MCB_X_SIZE(p)+=MCB_X_SIZE(MCB_NEXT(p))+MCB_PSIZE;
#ifdef DEBUG_MEMORY_FUNCTION_PROBLEMS
  if (stddebug) fprintf(stddebug,"join blocks leave\n");
#endif
}



/* Allocate DOS memory. Or'ed SEARCH_NOSPLIT with mode parameteres casues not
to split memory block but allocate the whole block for us */
int dosmem_alloc ( int size , int mode , int user )
{
  byte *p=dosmem_search(size,mode&SEARCH_MASK);
  int nosplit=mode&SEARCH_NOSPLIT;
#ifdef DEBUG_MEMORY_FUNCTION_PROBLEMS
  if (stddebug) fprintf(stddebug,"dosmem_alloc enter, requested size : %04X\n",size);
#endif
  if (p==NULL) return -DNOMEM; /* not enough memory ? */
  if (MCB_X_SIZE(p)-size>=MCB_MAX_FRAGMENT&&!nosplit) {
    byte *q=split_block(p,size);
    if (MCB_X_TYPE(q)=='M'&&!MCB_X_OWNER(MCB_NEXT(q))) join_blocks(q);
  }
  MCB_X_OWNER(p)=user;
#ifdef DEBUG_MEMORY_FUNCTION_PROBLEMS
  dosmem_dump();
  if (stddebug) fprintf(stddebug,"dosmem_alloc leave : %04X\n",((int)p+MCB_SIZE)>>4);
#endif
  return (int)p+MCB_SIZE;
}



int dosmem_free ( int addr , int user )
{
  byte *p=(byte*)(addr-MCB_SIZE);
#ifdef DEBUG_MEMORY_FUNCTION_PROBLEMS
  if (stddebug) fprintf(stddebug,"dosmem_free enter : %04X\n",addr>>4);
#endif
  if (addr<=MCB_SIZE) return DBADMCB;
  if (MCB_X_TYPE(p)!='M'&&MCB_X_TYPE(p)!='Z') {
    if (stddebug) fprintf(stddebug,"Trying free non MCB struct !!\n");
    return DBADMCB;
  }
  if (MCB_X_OWNER(p)!=user) return 1;
  MCB_X_OWNER(p)=0;
#ifdef DEBUG_MEMORY_FUNCTION_PROBLEMS
  if (stddebug) fprintf(stddebug,"dosmem_free leave\n");
#endif
  return optimize_blocks();
}



int dosmem_realloc ( int addr , int newsize , int user )
{
  byte *p=(byte*)(addr-MCB_SIZE);
  int remain;
#ifdef DEBUG_MEMORY_FUNCTION_PROBLEMS
  if (stddebug) fprintf(stddebug,"dosmem_realloc enter : addr=%04X size=%04X newsize=%04X\n",addr>>4,MCB_X_SIZE(p),newsize);
  dosmem_dump();
#endif
  if (MCB_X_TYPE(p)!='M'&&MCB_X_TYPE(p)!='Z') {
    if (stddebug) fprintf(stddebug,"Trying realloc non-MCB struct !!\n");
    return DBADMCB;
  } /*
  if (MCB_X_OWNER(p)!=user) {
    if (stddebug) fprintf(stddebug,"Trying realloc non owned block !!\n");
    return DBADMCB;
  } */
  if (newsize==MCB_X_SIZE(p)) return 0; /* actual size */
  if (newsize>MCB_X_SIZE(p)) {  /* BLOCK GROW   */
#ifdef DEBUG_MEMORY_FUNCTION_PROBLEMS
    if (stddebug) fprintf(stddebug,"grow\n");
#endif
    if (MCB_X_TYPE(p)=='Z'||MCB_X_OWNER(MCB_NEXT(p))) return DNOMEM; /* no free block after us */
    remain=MCB_X_SIZE(p)+MCB_X_SIZE(MCB_NEXT(p))+MCB_PSIZE-newsize;
    if (remain<0) return DNOMEM; /* not enough memory in our two blocks */
    join_blocks(p);
    if (remain>=MCB_MAX_FRAGMENT) (void)split_block(p,newsize);
  } else {                /* BLOCK SHRINK */
#ifdef DEBUG_MEMORY_FUNCTION_PROBLEMS
    if (stddebug) fprintf(stddebug,"shrink\n");
#endif
    if (MCB_X_SIZE(p)-newsize>MCB_MAX_FRAGMENT) {
      byte *q=split_block(p,newsize);
      if (MCB_X_TYPE(q)=='M'&&!MCB_X_OWNER(MCB_NEXT(q))) join_blocks(q);
    }
  }
#ifdef DEBUG_MEMORY_FUNCTION_PROBLEMS
  dosmem_dump();
  if (stddebug) fprintf(stddebug,"dosmem_realloc leave\n");
#endif
  return 0;
}



void dosmem_dump ( void )
{
  if (stddebug) {
  byte *p=mcb_first;
  int a;
#ifdef DEBUG_MEMORY_FUNCTION_PROBLEMS
  fprintf(stddebug,"dosmem_dump enter\n");
#endif
  fprintf(stddebug,"MCB  data size owns type hex-byte-dump-of-MCB...\n");
  for (;;) {
    byte *b=(byte*)p;
    fprintf(stddebug,"%04X %04X %04X %04X %c   ",(int)p>>4,((int)p+MCB_SIZE)>>4,MCB_X_SIZE(p),MCB_X_OWNER(p),MCB_X_TYPE(p));
    for (a=0;a<16;a++) fprintf(stddebug," %02X",*(b++));
    fprintf(stddebug,"\n");
    if (MCB_X_TYPE(p)!='M') break; else p=MCB_NEXT(p);
  }
  if (MCB_X_TYPE(p)!='Z') fprintf(stddebug,"dosmem_dump: MCB list corrupted (dosmem_dump) !!\n");
  fprintf(stddebug,"%04X ------- END -------\n",((int)MCB_NEXT(p))>>4);
#ifdef DEBUG_MEMORY_FUNCTION_PROBLEMS
  fprintf(stddebug,"dosmem_dump leave\n");
#endif
  }
}
